@@ -4,10 +4,10 @@
  * ANSI/POSIX + Microsoft compatible standard library function prototypes,
  * associated macros, and manifest constant definitions.
  *
- * $Id: stdlib.h,v 730a8ef64a92 2018/12/20 19:30:25 keith $
+ * $Id: stdlib.h,v 1aa8433dd381 2020/03/04 19:20:34 keith $
  *
  * Written by Colin Peters <colin@bird.fu.is.saga-u.ac.jp>
- * Copyright (C) 1997-2009, 2011, 2014-2016, 2018, MinGW.org Project.
+ * Copyright (C) 1997-2009, 2011, 2014-2016, 2018, 2020, MinGW.org Project.
  *
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -156,16 +156,43 @@ __MINGW_IMPORT char **__argv_dll;
 #endif  /*  __DECLSPEC_SUPPORTED */
 #endif  /* MB_CUR_MAX */
 
-/* FIXME: Nominally in <errno.h>, Microsoft likes to declare errno
- * in <stdlib.h> as well; we should factor this out.
+/* In MSVCR80.DLL, Microsoft introduced the following pair of errno
+ * accessor functions; they subsequently became available in MSVCRT.DLL
+ * from Vista onward.  Although they are not required by ISO-C, and they
+ * are more cumbersome to use, than referring to errno directly, the GCC
+ * developers have gratuitously chosen to assume, in GCC-9.x, that they
+ * are always supported on MS-Windows, regardless of Windows version.
+ * Logically, we might expect these to be declared in <errno.h>, but
+ * Microsoft's documentation insists that they are actually declared
+ * here; thus, to satisfy the GCC-9.x requirement, we will declare
+ * them unconditionally here ...
  */
-#ifdef _UWIN
-# undef errno
-  extern int errno;
-#else
-_CRTIMP __cdecl __MINGW_NOTHROW  int *_errno(void);
-# define errno  (*_errno())
+__cdecl __MINGW_NOTHROW  int _get_errno(int *);
+__cdecl __MINGW_NOTHROW  int _set_errno(int);
+
+/* ... then provide in-line implementations, (depending on gratuitous
+ * exposure of EINVAL, which strictly belongs in <errno.h> only, while
+ * also requiring declaring the ISO-C errno feature, which Microsoft
+ * documentation calls for both here, and in <errno.h>; we satisfy
+ * both of these requirements by selective <errno.h> inclusion).
+ */
+#define __STDLIB_H_SOURCED__ 1
+#include "errno.h"
+
+#if __MSVCRT_VERSION__ < __MSVCR80_DLL && _WIN32_WINNT < _WIN32_WINNT_VISTA
+/* These in-line implementations will support universal use of this API,
+ * even on legacy Windows versions pre-dating Vista, without requiring use
+ * of non-free MSVCRT80.DLL or later.
+ */
+__CRT_ALIAS __cdecl __MINGW_NOTHROW  int _get_errno( int *__val )
+{ return (__val == NULL) ? (errno = EINVAL) : 0 & (*__val = errno); }
+
+__CRT_ALIAS __cdecl __MINGW_NOTHROW  int _set_errno( int __val )
+{ errno = __val; return 0; }
+
 #endif
+#undef __STDLIB_H_SOURCED__
+
 _CRTIMP __cdecl __MINGW_NOTHROW  int *__doserrno(void);
 #define _doserrno  (*__doserrno())
 
@@ -399,6 +426,20 @@ unsigned long wcstoul (const wchar_t *, wchar_t **, int);
 
 _CRTIMP __cdecl __MINGW_NOTHROW  double wcstod (const wchar_t *, wchar_t **);
 
+/* The following MinGW specific alternatives to wcstod(), which may
+ * offer more robust performance than the MSVCRT.DLL implementation,
+ * are provided in libmingwex.a; (the float and long double variants
+ * are simply aliases for the ISO-C99 equivalents which follow).
+ */
+__cdecl __MINGW_NOTHROW
+double __mingw_wcstod (const wchar_t *__restrict__, wchar_t **__restrict__);
+
+__cdecl __MINGW_NOTHROW
+float __mingw_wcstof (const wchar_t *__restrict__, wchar_t **__restrict__);
+
+__cdecl __MINGW_NOTHROW
+long double __mingw_wcstold (const wchar_t *__restrict__, wchar_t **__restrict__);
+
 #ifdef _ISOC99_SOURCE
 /* Variants on wcstod(), specified by ISO-C99; once again, MSVCRT.DLL
  * doesn't have them, but we offer them in libmingwex.a
@@ -481,6 +522,25 @@ _CRTIMP __cdecl __MINGW_NOTHROW  int mbtowc (wchar_t *, const char *, size_t);
 _CRTIMP __cdecl __MINGW_NOTHROW  int rand (void);
 _CRTIMP __cdecl __MINGW_NOTHROW  void srand (unsigned int);
 
+/* rand() is devoid of entropy, and is thus a mediocre pseudo-random number
+ * generator.  Microsoft do offer a better quality (bogusly dubbed as a more
+ * secure) PRNG, in the guise of rand_s(), but it
+ *
+ *   1) must be explicitly enabled, by user defined feature test macro;
+ */
+#ifdef _CRT_RAND_S
+/*
+ *   2) is not supported on Win9x, nor any WinNT version prior to WinXP;
+ *   3) on WinXP, requires linking with non-free MSVCR80.DLL, or later;
+ *   4) is provided by MSVCRT.DLL, only from Vista onward.
+ */
+#if __MSVCRT_VERSION__ >= __MSVCR80_DLL || _WIN32_WINNT >= _WIN32_WINNT_VISTA
+
+_CRTIMP __cdecl __MINGW_NOTHROW  int rand_s (unsigned int *);
+
+#endif	/* Win-Vista || MSVCR80.DLL || later */
+#endif	/* _CRT_RAND_S enabled */
+
 _CRTIMP __cdecl __MINGW_NOTHROW  void abort (void) __MINGW_ATTRIB_NORETURN;
 _CRTIMP __cdecl __MINGW_NOTHROW  void exit (int) __MINGW_ATTRIB_NORETURN;
 
